@@ -87,7 +87,9 @@ WindowPreview.prototype = {
         this.actor = new St.Bin({style_class: "switcher-list", style: "margin: 0px; padding: 8px;"});
         this.actor.show_on_set_parent = false;
 
-        this.actor.set_size(WINDOW_PREVIEW_WIDTH * 1.3, WINDOW_PREVIEW_HEIGHT * 1.3);
+        this.scaleFactor = St.ThemeContext.get_for_stage(global.stage).scale_factor;
+
+        this.actor.set_size(WINDOW_PREVIEW_WIDTH * 1.3 * this.scaleFactor, WINDOW_PREVIEW_HEIGHT * 1.3 * this.scaleFactor);
         Main.uiGroup.add_actor(this.actor);
 
         this.metaWindow = metaWindow;
@@ -102,10 +104,9 @@ WindowPreview.prototype = {
         iconBox.set_child(icon);
         hbox.add_actor(iconBox);
 
-        let label = new St.Label();
-        label.set_text(this.metaWindow.get_title());
-        label.style = "padding: 2px;";
-        hbox.add_actor(label);
+        this.label = new St.Label();
+        this.label.style = "padding: 2px;";
+        hbox.add_actor(this.label);
 
         box.add_actor(hbox);
 
@@ -130,7 +131,7 @@ WindowPreview.prototype = {
     },
 
     show: function() {
-        if (!this.actor)
+        if (!this.actor || this._applet._menuOpen)
             return
 
         let muffinWindow = this.metaWindow.get_compositor_private();
@@ -145,8 +146,8 @@ WindowPreview.prototype = {
 
         this.thumbnail = new Clutter.Clone({
             source: windowTexture,
-            width: width * scale,
-            height: height * scale
+            width: width * scale * this.scaleFactor,
+            height: height * scale * this.scaleFactor
         });
 
         this.thumbnailBin.set_child(this.thumbnail);
@@ -185,6 +186,10 @@ WindowPreview.prototype = {
         this.visible = false;
     },
 
+    set_text: function(text) {
+        this.label.set_text(text);
+    },
+
     _destroy: function() {
         if (this.thumbnail) {
             this.thumbnailBin.set_child(null);
@@ -278,12 +283,12 @@ AppMenuButton.prototype = {
         if (this._tooltip)
             this._tooltip.destroy();
 
-        if (this._applet.usePreview) {
+        if (this._applet.usePreview)
             this._tooltip = new WindowPreview(this, this.metaWindow, this._applet.orientation);
-        } else {
+        else
             this._tooltip = new Tooltips.PanelItemTooltip(this, "", this._applet.orientation);
-            this.setDisplayTitle();
-        }
+        
+        this.setDisplayTitle();
     },
 
     onPanelEditModeChanged: function() {
@@ -400,6 +405,9 @@ AppMenuButton.prototype = {
         if (title.length > MAX_TEXT_LENGTH)
             title = title.substr(0, MAX_TEXT_LENGTH);
 
+        if (this._tooltip  && this._tooltip.set_text)
+            this._tooltip.set_text(title);
+
         if (this.metaWindow.minimized) {
             title = "["+ title +"]";
         } else if (this.metaWindow.tile_type == Meta.WindowTileType.TILED) {
@@ -410,8 +418,6 @@ AppMenuButton.prototype = {
         }
 
         this._label.set_text(title);
-        if (this._tooltip  && this._tooltip.set_text)
-            this._tooltip.set_text(title);
     },
 
     destroy: function() {
@@ -791,6 +797,11 @@ AppMenuButtonRightClickMenu.prototype = {
     },
 
     _onToggled: function(actor, isOpening){
+        if (this.isOpen)
+            this._launcher._applet._menuOpen = true;
+        else
+            this._launcher._applet._menuOpen = false;
+
         if (!isOpening) {
             return;
         }
@@ -816,6 +827,7 @@ MyApplet.prototype = {
         this.dragInProgress = false;
         this._tooltipShowing = false;
         this._tooltipErodeTimer = null;
+        this._menuOpen = false;
         this._urgentSignal = null;
         this._windows = [];
         this._monitorWatchList = [];
@@ -853,11 +865,12 @@ MyApplet.prototype = {
 
         let tracker = Cinnamon.WindowTracker.get_default();
         this.signals.connect(tracker, "notify::focus-app", this._onFocus);
-        this.signals.connect(global.screen, 'window-entered-monitor', this._onWindowAdded);
-        this.signals.connect(global.screen, 'window-left-monitor', this._onWindowRemoved);
-        this.signals.connect(global.screen, 'notify::n-workspaces', this._onNWorkspacesChanged);
+        this.signals.connect(global.screen, 'window-added', this._onWindowAdded);
+        this.signals.connect(global.screen, 'window-removed', this._onWindowRemoved);
+        this.signals.connect(global.screen, 'window-monitor-changed', this._onWindowMonitorChanged);
+        this.signals.connect(global.screen, 'window-workspace-changed', this._onWindowWorkspaceChanged);
         this.signals.connect(global.screen, 'monitors-changed', this._updateWatchedMonitors);
-        this.signals.connect(global.window_manager, 'switch-workspace',this._refreshItems);
+        this.signals.connect(global.window_manager, 'switch-workspace', this._refreshAllItems);
 
         this.signals.connect(global.window_manager, 'minimize', this._onWindowStateChange);
         this.signals.connect(global.window_manager, 'maximize', this._onWindowStateChange);
@@ -868,7 +881,6 @@ MyApplet.prototype = {
         global.settings.bind("panel-edit-mode", this.actor, "reactive", Gio.SettingsBindFlags.DEFAULT);
 
         this.on_orientation_changed(orientation);
-        this._onNWorkspacesChanged();
         this._updateAttentionGrabber();
     },
 
@@ -881,7 +893,7 @@ MyApplet.prototype = {
     },
 
     on_panel_height_changed: function() {
-        this._refreshItems();
+        this._refreshAllItems();
     },
 
     on_orientation_changed: function(orientation) {
@@ -901,27 +913,27 @@ MyApplet.prototype = {
         }
     },
 
-    /* These are callbacks to two different signals that pass different number
-     * of arguments. However, the last argument is always metaWindow, which is
-     * the only thing we want. */
-    _onWindowAdded: function() {
-        if (this._shouldAdd(arguments[arguments.length - 1]))
-            this._addWindow(arguments[arguments.length - 1]);
+    _onWindowAdded: function(screen, metaWindow, monitor) {
+        if (this._shouldAdd(metaWindow))
+            this._addWindow(metaWindow);
     },
 
-    _onWindowRemoved: function() {
-        this._removeWindow(arguments[arguments.length - 1]);
+    _onWindowRemoved: function(screen, metaWindow) {
+        this._removeWindow(metaWindow);
     },
 
-    _onNWorkspacesChanged: function() {
-        this.signals.disconnect("window-added");
-        this.signals.disconnect("window-removed");
+    _onWindowMonitorChanged: function(screen, metaWindow, monitor) {
+        if (this._shouldAdd(metaWindow))
+            this._addWindow(metaWindow);
+        else
+            this._removeWindow(metaWindow);
+    },
 
-        for (let i = 0; i < global.screen.n_workspaces; i++) {
-            let ws = global.screen.get_workspace_by_index(i);
-            this.signals.connect(ws, "window-added", this._onWindowAdded);
-            this.signals.connect(ws, "window-removed", this._onWindowRemoved);
-        }
+    _onWindowWorkspaceChanged: function(screen, metaWindow, metaWorkspace) {
+        let window = this._windows.find(win => (win.metaWindow == metaWindow));
+
+        if (window)
+            this._refreshItem(window);
     },
 
     _updateAttentionGrabber: function() {
@@ -966,18 +978,22 @@ MyApplet.prototype = {
             window.onFocus();
     },
 
-    _refreshItems: function() {
-        for (let window of this._windows) {
-            window.actor.visible =
-                (window.metaWindow.get_workspace() == global.screen.get_active_workspace()) ||
-                window.metaWindow.is_on_all_workspaces();
+    _refreshItem: function(window) {
+        window.actor.visible =
+            (window.metaWindow.get_workspace() == global.screen.get_active_workspace()) ||
+            window.metaWindow.is_on_all_workspaces();
 
-            /* The above calculates the visibility if it were the normal
-             * AppMenuButton. If this is actually a temporary AppMenuButton for
-             * urgent windows on other workspaces, it is shown iff the normal
-             * one isn't shown! */
-            if (window.alert)
-                window.actor.visible = !window.actor.visible;
+        /* The above calculates the visibility if it were the normal
+         * AppMenuButton. If this is actually a temporary AppMenuButton for
+         * urgent windows on other workspaces, it is shown iff the normal
+         * one isn't shown! */
+        if (window.alert)
+            window.actor.visible = !window.actor.visible;
+    },
+
+    _refreshAllItems: function() {
+        for (let window of this._windows) {
+            this._refreshItem(window);
         }
         this._onFocus();
     },
@@ -1051,7 +1067,7 @@ MyApplet.prototype = {
          * default, so move it to the start if needed */
         if (alert) {
             if (metaWindow.get_workspace().index() < global.screen.get_active_workspace_index())
-                this.actor.move_child(appButton.actor, 0);
+                this.actor.set_child_at_index(appButton.actor, 0);
         } else {
             if (metaWindow.get_workspace() != global.screen.get_active_workspace())
                 appButton.actor.hide();
@@ -1092,10 +1108,10 @@ MyApplet.prototype = {
             this._dragPlaceholder = new DND.GenericDragPlaceholderItem();
             this._dragPlaceholder.child.set_width (source.actor.width);
             this._dragPlaceholder.child.set_height (source.actor.height);
-            this.actor.insert_actor(this._dragPlaceholder.actor,
-                    this._dragPlaceholderPos);
+            this.actor.insert_child_at_index(this._dragPlaceholder.actor,
+                                             this._dragPlaceholderPos);
         } else {
-            this.actor.move_child(this._dragPlaceholder.actor,
+            this.actor.set_child_at_index(this._dragPlaceholder.actor,
                     this._dragPlaceholderPos);
         }
 
@@ -1106,7 +1122,7 @@ MyApplet.prototype = {
         if (!(source instanceof AppMenuButton)) return false;
         if (this._dragPlaceholderPos == undefined) return false;
 
-        this.actor.move_child(source.actor, this._dragPlaceholderPos);
+        this.actor.set_child_at_index(source.actor, this._dragPlaceholderPos);
 
         return true;
     },
